#include <stdio.h>
#include <git2.h>
#include <rpmlib.h>
#include <argp.h>
#include <sys/queue.h>
#include <strings.h>
#include <regex.h>

#include "c-vector/vec.h"

const char * argp_program_version = "0.1";
const char * argp_program_bug_address = "<clime@fedoraproject.org>";
static char doc[] =
    "For TARGET_REPO, outputs N-V-R named annotated tags of the given PKG_NAME (N part of N-V-R) that are "
    "reachable from the current HEAD or the \"start-commit\" argument if given, sorted primarily in topological order.\n"
    "Tags which are topologically unsortable will be sorted by a secondary criterion which is native "
    "rpm version and release sorting.";
static char args_doc[] = "TARGET_REPO PKG_NAME";
static struct argp_option options[] = {
    { "start-commit", 's', "COMMITTISH", 0, "Starting point for commit traversal. By default HEAD." },
    { 0 }
};

git_repository * repo = NULL;

typedef struct {
    char * start_commit;
    char * target_repo;
    char * name;
} Arguments;

typedef struct {
    char * name;
    git_oid oid;
    git_tag * tag;
} Tag;

typedef struct  {
    struct result_entry * e;
    int e_idx;
} InsertedResult;

typedef struct  {
    int max_e_idx;
    struct result_entry * max_e;
    git_oid oid;
} Visited;

Tag * tags;

/* we don't use typedef here because 
 * TAILQ macros assume "struct result_entry" */
struct result_entry {
    int tag_idx;
    TAILQ_ENTRY(result_entry) entries;
};

TAILQ_HEAD(tailqhead, result_entry) result_head = TAILQ_HEAD_INITIALIZER(result_head);

/*
 * Parse a single option from command-line.
 */
static error_t parse_opt(int key, char * arg, struct argp_state * state) {
    /* Get the input argument from argp_parse, which we
       know is a pointer to our arguments structure. */
    Arguments * arguments = state->input;

    switch (key)
    {
        case 's':
            arguments->start_commit = arg;
            break;

        case ARGP_KEY_ARG:
            if (state->arg_num == 0)
                arguments->target_repo = arg;
            else if (state->arg_num == 1)
                arguments->name = arg;
            else /* Too many arguments. */
                argp_usage(state);
            break;

        case ARGP_KEY_END:
            if (state->arg_num < 2)
                /* Not enough arguments. */
                argp_usage(state);
            break;

        default:
            return ARGP_ERR_UNKNOWN;
    }
    return 0;
}

/* 
 * return true if tag1_name <= tag2_name
 * according to rpm version sort, false otherwise.
 *
 * We assume the N part of tag names (composed of N-V-R)
 * is the same for both tags and we compare only
 * V (primary criterion) and R (secondary one).
 */
bool rpm_is_lower_than(char * tag1_name, char * tag2_name) {
    regex_t re;
    int status;
    char * pattern = "^.+-([^-]+)-([^-]+)$";

    if (regcomp(&re, pattern, REG_EXTENDED) != 0) {
        fprintf(stderr, "Could not compile regular expression.\n");
        exit(1);
    }

    size_t ngroups = re.re_nsub + 1;
    regmatch_t * groups1 = malloc(ngroups * sizeof(regmatch_t));
    regmatch_t * groups2 = malloc(ngroups * sizeof(regmatch_t));

    status = regexec(&re, tag1_name, ngroups, groups1, 0);
    if (status != 0) {
        fprintf(stderr, "Tag name %s is not in expected N-V-R form.\n", tag1_name);
        exit(1);
    }

    status = regexec(&re, tag2_name, ngroups, groups2, 0);
    if (status != 0) {
        fprintf(stderr, "Tag name %s is not in expected N-V-R form.\n", tag2_name);
        exit(1);
    }
    regfree(&re);

    /* copy tag names */
    char tag1_name_cpy[strlen(tag1_name) + 1];
    char tag2_name_cpy[strlen(tag1_name) + 1];
    strcpy(tag1_name_cpy, tag1_name);
    strcpy(tag2_name_cpy, tag2_name);

    /* sanity check on matches */
    if (groups1[1].rm_so == (size_t)-1 || groups1[2].rm_so == (size_t)-1 ||
            groups2[1].rm_so == (size_t)-1 || groups2[2].rm_so == (size_t)-1) {
        fprintf(stderr, "Did not get expected matches during rpm version comparison.\n");
        exit(1);
    }

    /* place zero markers into the copies according to the matches */
    tag1_name_cpy[groups1[1].rm_eo] = 0;
    tag1_name_cpy[groups1[2].rm_eo] = 0;
    tag2_name_cpy[groups2[1].rm_eo] = 0;
    tag2_name_cpy[groups2[2].rm_eo] = 0;

    /* extract versions and releases for both tags */
    char * tag1_version = malloc(groups1[1].rm_eo - groups1[1].rm_so + 1);
    char * tag1_release = malloc(groups1[2].rm_eo - groups1[2].rm_so + 1);
    char * tag2_version = malloc(groups2[1].rm_eo - groups2[1].rm_so + 1);
    char * tag2_release = malloc(groups2[2].rm_eo - groups2[2].rm_so + 1);

    strcpy(tag1_version, &tag1_name_cpy[groups1[1].rm_so]);
    strcpy(tag1_release, &tag1_name_cpy[groups1[2].rm_so]);
    strcpy(tag2_version, &tag2_name_cpy[groups2[1].rm_so]);
    strcpy(tag2_release, &tag2_name_cpy[groups2[2].rm_so]);

    /* do version comparison */
    int vercmp = rpmvercmp(tag1_version, tag2_version);
    if (vercmp < 0) {
        /* tag1's version is "older/lower" */
        return true;
    } else if (vercmp > 0) {
        /* tag1's version is "newer/higher" */
        return false;
    }

    /* versions were equal, do release comparison */
    int relcmp = rpmvercmp(tag1_release, tag2_release);
    if (relcmp < 0) {
        /* tag1's release is "older/lower" */
        return true;
    } else if (relcmp > 0) {
        /* tag1's release is "newer/higher" */
        return false;
    }

    /* both tags are equal, should not happen */
    return false;
}

/*
 * Insert tag (index) into sorted result list.
 *
 * e specifies where to insert the new tag_idx given that
 * there are no further records after e in the result. If
 * there are, new entry is inserted at the first location
 * where rpm version and release sort evaluates it as
 * lower than the next element, if any, otherwise at
 * the last position.
 *
 * Note that this will produce the same ordering as solely
 * rpm version sort would as long as tag NVRs raise with
 * topologically newer commits.
 *
 * e_idx is an accompanying index of e in the result
 * list.
 *
 * The function returns pointer to the newly inserted
 * element and its index in the result list.
 */
InsertedResult add_to_result(int tag_idx, struct result_entry * e, int e_idx) {
    struct result_entry * new_result_entry = malloc(sizeof(struct result_entry));
    new_result_entry->tag_idx = tag_idx;

    struct result_entry * next;
    if (!e) {
        next = TAILQ_FIRST(&result_head);
    } else {
        next = TAILQ_NEXT(e, entries);
    }

    InsertedResult ir;
    ir.e = new_result_entry;
    struct result_entry * current = e;
    for (; next != NULL; next = TAILQ_NEXT(next, entries)) {
        if (rpm_is_lower_than(tags[tag_idx].name, tags[next->tag_idx].name)) {
            if (!current) {
                TAILQ_INSERT_HEAD(&result_head, new_result_entry, entries);
            } else {
                TAILQ_INSERT_AFTER(&result_head, current, new_result_entry, entries);
            }
            ir.e_idx = e_idx + 1;
            return ir;
        }
        current = next;
        e_idx++;
    }

    /* next is NULL so insert after current
     * or insert head if current is NULL */
    if (!current) {
        TAILQ_INSERT_HEAD(&result_head, new_result_entry, entries);
        ir.e_idx = 0;
    } else {
        TAILQ_INSERT_AFTER(&result_head, current, new_result_entry, entries);
        ir.e_idx = e_idx + 1;
    }

    return ir;
}

Visited * is_oid_visited(const git_oid * oid, Visited * visited) {
    for (int i = 0; i < vector_size(visited); i++) {
        /* if git_oid_cmp returns 0, then the two oids are the same */
        if (git_oid_cmp(oid, &visited[i].oid) == 0) {
            return &visited[i];
        }
    }
    return NULL;
}

/* 
 * DFS on git history. As annotated tags from tags
 * are encountered, they are (their indeces to be
 * more precise) sorted into the resulting list.
 */
InsertedResult visit(git_repository * repo, git_commit * c, const char * name, Visited ** visited_ptr) {
    size_t i, num_parents = git_commit_parentcount(c);
    InsertedResult max_ir;
    max_ir.e = NULL;
    max_ir.e_idx = -1;

    for (i = 0; i < num_parents; i++) {
        git_commit * p;
        if (git_commit_parent(&p, c, i) == 0) {
            Visited * visited = is_oid_visited(git_commit_id(p), *visited_ptr);
            if (!visited) {
                InsertedResult ir = visit(repo, p, name, visited_ptr);
                if (ir.e_idx > max_ir.e_idx) {
                    max_ir.e = ir.e;
                    max_ir.e_idx = ir.e_idx;
                } else {
                    /* max element was shifted by the
                     * newly inserted element */
                    max_ir.e_idx++;
                }
            } else {
                /* inherit previously computed values
                 * for an already visited commit */
                max_ir.e = visited->max_e;
                max_ir.e_idx = visited->max_e_idx;
            }
        }
        git_commit_free(p);
    }

    /* check if the commit has an attached tag from tags */
    int * tag_indeces_found = vector_create();
    for (int i = 0; i < vector_size(tags); i++) {
        const git_oid * tag_commit_oid = git_tag_target_id(tags[i].tag);
        const git_oid * current_commit_oid = git_commit_id(c);
        /* if git_oid_cmp returns 0, then the two oids are the same */
        if (git_oid_cmp(tag_commit_oid, current_commit_oid) == 0) {
            vector_add(&tag_indeces_found, i);
        }
    }

    struct result_entry * bound_e = max_ir.e;
    int bound_e_idx = max_ir.e_idx;
    for (int i = 0; i < vector_size(tag_indeces_found); i++) {
        InsertedResult ir = add_to_result(tag_indeces_found[i], bound_e, bound_e_idx);
        if (ir.e_idx > max_ir.e_idx) {
            max_ir.e = ir.e;
            max_ir.e_idx = ir.e_idx;
        } else {
            max_ir.e_idx++; 
        }
    }

    /* put current commit into visited */
    Visited * temp = vector_add_asg(visited_ptr);
    git_oid_cpy(&temp->oid, git_commit_id(c));
    temp->max_e_idx = max_ir.e_idx;
    temp->max_e = max_ir.e;

    vector_free(tag_indeces_found);
    return max_ir;
}

/*
 * Match string against the extended regular expression in
 * pattern, treating errors as no match.
 *
 * Return true for match, false for no match.
 */
bool match(const char * string, char * pattern) {
    int status;
    regex_t re;

    if (regcomp(&re, pattern, REG_EXTENDED|REG_NOSUB) != 0) {
        fprintf(stderr, "Could not compile regular expression.\n");
        exit(1);
    }

    status = regexec(&re, string, (size_t) 0, NULL, 0);
    regfree(&re);

    if (status != 0) {
        return false;
    }
    return true;
}

/* 
 * git_tag_foreach callback to populate tags.
 */
int each_tag(const char * path, git_oid * oid, void * payload) {
    // locate last "/"
    char * tag_name = rindex(path, '/');

    if (tag_name) {
        tag_name++;
    } else {
        tag_name = (char *) path;
    }

    char * name = (char *) payload;
    char * pattern = malloc(strlen(name)+15);
    strcpy(pattern, "^");
    strcat(pattern, name);
    strcat(pattern, "-[^-]+-[^-]+$");
    if (!match(tag_name, pattern)) {
        return 0; /* does not have the requested name */
    }

    git_tag * tag;
    if (git_tag_lookup(&tag, repo, oid)) {
        return 0; /* not an annotated tag */
    }

    Tag * tmp = vector_add_asg(&tags);
    tmp->name = malloc(strlen(tag_name)+1);
    strcpy(tmp->name, tag_name);
    git_oid_cpy(&tmp->oid, oid);
    tmp->tag = tag;
    return 0;
}

/* Our argp parser. */
static struct argp argp = { options, parse_opt, args_doc, doc };

int main(int argc, char **argv) {
    Arguments args;
    int err;

    /* Default values. */
    args.start_commit = "HEAD";

    /* Parse our arguments; every option seen by
     * parse_opt will be reflected in arguments. */
    argp_parse(&argp, argc, argv, 0, 0, &args);

    git_libgit2_init();

    /* open the requested repository */
    if ((err = git_repository_open(&repo, args.target_repo))) {
        const git_error * e = giterr_last();
        fprintf(stderr, "Error %d/%d: %s\n", err, e->klass, e->message);
        exit(err);
    }

    /* populate tags */
    tags = vector_create();
    if ((err = git_tag_foreach(repo, each_tag, args.name))) {
        const git_error * e = giterr_last();
        fprintf(stderr, "Error %d/%d: %s\n", err, e->klass, e->message);
        exit(err);
    }

    /* fetch start commit */
    git_object * obj;
    if ((err = git_revparse_single(&obj, repo, args.start_commit))) {
        const git_error * e = giterr_last();
        fprintf(stderr, "Error %d/%d: %s\n", err, e->klass, e->message);
        exit(err);
    }

    /* try casting to git commit */
    git_commit * start_commit;
    if (git_object_type(obj) == GIT_OBJ_COMMIT) {
        start_commit = (git_commit *) obj;
    } else {
        fprintf(stderr, "start-commit argument gave a non-commit.\n");
        exit(1);
    }

    /* run DFS on git history */
    Visited * visited = vector_create();
    visit(repo, start_commit, args.name, &visited);
    vector_free(visited);

    /* print the sorted tags */
    char sha[41] = { 0 };
    struct result_entry * e;
    for (e = TAILQ_LAST(&result_head, tailqhead); e != NULL; e = TAILQ_PREV(e, tailqhead, entries)) {
        git_oid_tostr(sha, 41, &tags[e->tag_idx].oid);
        printf("%s %s\n", sha, tags[e->tag_idx].name);
    }

    return 0;
}
